# -*- coding: utf-8 -*-
"""
compute joint statistics and generate plots for a given set of phenotypes
"""
from __future__ import absolute_import
from typing import List
import os
import hashlib

from jass.models.base_model_ import Model
from jass.util import deserialize_model
from jass.models.phenotype import Phenotype
from jass.models.worktable import (
    get_worktable_summary,
    get_worktable_genomedata,
    get_worktable_local_manhattan_data,
    get_worktable_local_heatmap_data,
)

from jass.config import config


class Project(Model):

    DOES_NOT_EXIST = "DOES_NOT_EXIST"

    CREATING = "CREATING"

    READY = "READY"

    ERROR = "ERROR"

    # Type of analysis
    LOCAL_ANALYSIS = "LOCAL_ANALYSIS"
    GENOME_WIDE_ANALYSIS = "GENOME_WIDE_ANALYSIS"

    # Upper bound of the chromosome length (bp)
    K_POS_MAX = 250000000

    def __init__(
        self,
        id: str = None,
        phenotypes: List[Phenotype] = None,
        chromosome: str = None,
        start: str = None,
        end: str = None,
    ):
        """
        Project - a project (list of phenotypes)

        :param id: project ID.
        :type id: str
        """
        self.swagger_types = {
            "id": str,
            "status": str,
            "phenotypes": List[Phenotype],
            "progress": str,
        }

        self.attribute_map = {
            "id": "id",
            "status": "status",
            "phenotypes": "phenotypes",
            "progress": "progress",
        }

        self._phenotypes = phenotypes

        self._id = id

        if self._id is None:
            self._id = self.get_id(chromosome, start, end)

    @classmethod
    def from_dict(cls, dikt) -> "Project":
        """
        Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Project.
        :rtype: Project
        """
        return deserialize_model(dikt, cls)

    @property
    def id(self) -> str:
        """
        Gets the id of this Project.

        :return: The id of this Project.
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id: str):
        """
        Lists the id of this Project.

        :param id: The id of this Project.
        :type id: str
        """

        self._id = id

    @property
    def phenotypes(self) -> List[Phenotype]:
        """
        Gets the phenotypes list for this project.

        :return: The phenotypes.
        :rtype: str
        """
        return self._phenotypes

    @phenotypes.setter
    def cohort(self, phenotypes: List[Phenotype]):
        """
        Lists the phenotypes list for this project.

        :param phenotypes: The phenotypes.
        :type phenotypes: str
        """

        self._phenotypes = phenotypes

    def get_type_of_analysis(self):
        """
        get_type_of_analysis
        Gets the type of analysis : local or genome wide
        """
        if (self._id).split("_")[0] == "local":
            return Project.LOCAL_ANALYSIS
        else:
            return Project.GENOME_WIDE_ANALYSIS

    def get_folder_path(self):
        """
        get_folder_path
        Gets the path of the folder where the project data are stored
        """
        return os.path.join(config["DATA_DIR"], "project_{}".format(self.id))

    def get_worktable_path(self):
        """
        get_worktable_path
        Gets the path of the file workTable.hdf5
        """
        return os.path.join(self.get_folder_path(), "workTable.hdf5")

    def get_csv_path(self):
        """
        get_csv_path
        Gets the path of the file genome_full.csv
        """
        return os.path.join(self.get_folder_path(), "workTable.csv")

    def get_progress_path(self):
        """
        get_progress_path
        Gets the path of the file containing the current progress percentage
        of the analysis performed within the project
        """
        return os.path.join(self.get_folder_path(), "JASS_progress.txt")

    def get_csv_lock_path(self):
        """
        get_csv_lock_path
        Gets the path of the lock set-on when the csv file is not available yet
        """
        return os.path.join(self.get_folder_path(), "the_lock.txt")

    def get_project_summary_statistics(self):
        return get_worktable_summary(self.get_worktable_path())

    def get_project_genomedata(self):
        return get_worktable_genomedata(self.get_worktable_path())

    def get_project_local_manhattan_data(
        self, chromosome: str = None, region: str = None
    ):
        print("project::get_project_local_manhattan_data")
        return get_worktable_local_manhattan_data(
            self.get_worktable_path(), chromosome, region
        )

    def get_project_local_heatmap_data(
        self, chromosome: str = None, region: str = None
    ):
        return get_worktable_local_heatmap_data(
            self.get_worktable_path(), chromosome, region
        )

    def get_id(self, chromosome=None, start=None, end=None):
        m = hashlib.md5()
        for phenotype_id in [phenotype.id for phenotype in self._phenotypes]:
            m.update(str(phenotype_id).encode("utf-8"))

        if chromosome is not None:
            # Local analysis
            if start is None:
                start = 0
            if end is None:
                end = Project.K_POS_MAX
            Loc_An = "_{}_{}_{}".format(chromosome, start, end)
            m.update(str(Loc_An).encode("utf-8"))
            id_project = "local_{}".format(m.hexdigest())
        else:
            id_project = m.hexdigest()

        return id_project

    def get_global_manhattan_plot_path(self):
        return os.path.join(self.get_folder_path(), "Manhattan_Plot_Omnibus.png")

    def get_quadrant_plot_path(self):
        return os.path.join(self.get_folder_path(), "Quadrant_Plot_Omnibus.png")

    def get_zoom_plot_path(self):
        return os.path.join(self.get_folder_path(), "Zoom_Plot_Omnibus.png")

    @property
    def status(self):
        """
        status
        Gets the status of the project
        """
        if not os.path.exists(self.get_folder_path()):
            return Project.DOES_NOT_EXIST
        else:
            if self.get_type_of_analysis() == Project.LOCAL_ANALYSIS:
                Return_status = get_file_status(self.get_zoom_plot_path())
            else:
                Return_status = get_file_status(self.get_quadrant_plot_path())

            return {
                # WARNING: project status is hacked so that everything is ready
                # only once the final step has completed.
                # This avoids the apparent "corrupted hdf5" file situation
                "worktable": Return_status,
                "global_manhattan": Return_status,
                "quadrant_plot_status": Return_status,
                "zoom_plot": Return_status,
            }

    @property
    def progress(self):
        """
        progress
        Gets the percentage of completion of the phenotype analysis
        """
        JASS_progress = 0
        progress_path = self.get_progress_path()
        if os.path.exists(progress_path):
            file_progress = open(progress_path, "r")
            JASS_progress = file_progress.read()
            file_progress.close()
        return JASS_progress

    def get_csv_file_generation(self):
        """
        csv_file_generation
        Gets the status of the genome_full csv file generation
        """
        the_lock_path = self.get_csv_lock_path()
        csv_file = self.get_csv_path()
        csv_file_status = Project.CREATING
        if not os.path.isfile(the_lock_path):
            if os.path.isfile(csv_file):
                csv_file_status = Project.READY
            else:
                csv_file_status = Project.ERROR
        print("csv_file_generation:csv_file_status={}".format(csv_file_status))
        return csv_file_status


def get_file_building_tb_path(file_path):
    return file_path + ".log"


def get_file_status(file_path):
    if os.path.exists(file_path):
        return Project.READY
    elif os.path.exists(get_file_building_tb_path(file_path)):
        return Project.ERROR
    else:
        return Project.CREATING
